#ifndef ArduinosInSpace_h
#define ArduinosInSpace_h

#ifndef Arduino_h
#include <Arduino.h>
#endif

#include "AISRequest.h"
#include "AISManagedInput.h"
#include "AISUtil.h"
#include "ArduinosInSpaceCommands.h"
#include "ArduinosInSpaceRequests.h"

typedef void (*commandCallback)(int command, int payload);

const int sendBufSize = 10;
const int recvBufSize = 16;

// We check and discard comands higher than this.
// TODO: Maybe a better way to do this?
const int maxCommand = 300;

class ObjectsInSpace : public ObjectsInSpaceInterface
{
 public:
  typedef void (ObjectsInSpace::*OISSerialCallback)(OISCommand command);

  /** Default constructor.
      @param serial The serial instance this instance will use to
      communicate with the game. Usually "Serial".
      @param numRequests The number of request channels we'll allocate.
      @param numCommandPins The number of input pins we'll manage. Note
      that this does not include commands specified with
      `registerCommand()`.
  */
  ObjectsInSpace(Stream &serial, int numRequests, int numCommandPins);
  /** Deprecated constructor.
      @param serial The serial instance this instance will use to
      communicate with the game. Usually "Serial".
      @param numRequests The number of request channels we'll allocate.
  */
  ObjectsInSpace(Stream &serial, int numRequests);

  /* Handshaking */
  /** Initialise the connection, and perform handshaking with the game.
  */
  void begin();

  /* Syncing */
  /** Register a client command. See the protocol documentation for a
      full list of possible commands.
      @param command The command constant to register.
  */
  int registerCommand(OISCommand command);
  /** Register a client command and bind it to an input pin.
      ArduinosInSpace will poll this pin, and whenever it changes
      from inactive to active will send the given client command.
      See the protocol documentation for a full list of possible commands.
      @param pin The pin to monitor. ArduinosInSpace will set the mode.
      @param activeCommand The command to send when the pin goes active
      @param mode This sets the pin mode, as well as what readings will
      be considered active. When set to INPUT, the pin is active when HIGH.
      When set to INPUT_PULLUP, the pin is active when LOW.
  */
  int registerTrigger(uint8_t pin, OISCommand activeCommand,
		      uint8_t mode=INPUT);
  /** Register two commands bound to an input pin.
      ArduinosInSpace will poll this pin, and send activeCommand when the
      pin changes from inactive to active, and inactiveCommand when the
      pin changes from active to inactive. See the protocol documentation
      for a full list of possible commands.
      @param pin The pin to monitor. ArduinosInSpace will set the mode.
      @param activeCommand The command to send when the pin goes active.
      @param inactiveCommand The command to send when the pin goes inactive.
      @param mode This sets the pin mode, as well as what readings will
      be considered active. When set to INPUT, the pin is active when HIGH.
      When set to INPUT_PULLUP, the pin is active when LOW.
  */
  int registerTrigger(uint8_t pin, OISCommand activeCommand,
		      OISCommand inactiveCommand, uint8_t mode=INPUT);
  /** Register two commands bound to an input pin in latching mode.
      ArduinosInSpace will poll this pin. The first time it switches
      from inactive to active the first command will be sent. The
      second time it switches from inactive to active the second command
      will be sent, and so on. See the protocol documentation for a full
      list of possible commands.
      @param pin The pin to monitor. ArduinosInSpace will set the mode.
      @param firstCommand The command to send first active change.
      @param secondCommand The command to send second active change.
      @param active Thsi sets the pin mode, as well as what readings will
      be considered active. When set to INPUT, the pin is active when HIGH.
      When set to INPUT_PULLUP, the pin is active when LOW.
  */
  int registerLatching(uint8_t pin, OISCommand firstCommand,
		       OISCommand secondCommand, uint8_t mode=INPUT);
  /** Register a request for boolean data. See the protocol documentation
      for a full list of possible data.
      @param command The data command to request.
      @param callback A callback function to handle the received data.
  */
  int registerBool(OISCommand command, boolCallback callback);
  /** Register a request for boolean data. See the protocol documentation
      for a full list of possible data. When this command is passed a pin
      number, the library will automatically turn the pin on or off
      depending on data received from the game.
      @param command The data command to request.
      @param pin A pin number to control.
  */
  int registerBool(OISCommand command, int pin, bool invert=false);
  /** Register a request for numeric data. See the protocol documentation
      for a full list of possible data.
      @param command The data command to request.
      @param callback A callback function to handle the received data.
      @param invert If set to true, the pin state is inverted
  */
  int registerInt(OISCommand command, intCallback callback);
  /** Register a request for floating point data. See the protocol
      documentation for a full list of possible data.
      @param command The data command to request.
      @param callback A callback function to handle the received data.
  */
  int registerFloat(OISCommand command, floatCallback callback);
  /** Register a request for boolean data. This method passes data
      from the game directly to the callback, without further
      processing.
      @param command The data to request.
      @param callback A callback function to handle the received data.
  */
  int registerBoolRaw(OISCommand command, rawCallback callback);
  /** Register a request for numeric data. This method passes data
      from the game directly to the callback, without further
      processing.
      @param command The data to request.
      @param callback A callback function to handle the received data.
  */
  int registerIntRaw(OISCommand command, rawCallback callback);
  /** Register a request for floating point data. This method passes
      data from the game directly to the callback, without further
      processing.
      @param command The data to request.
      @param callback A callback function to handle the received data.
  */
  int registerFloatRaw(OISCommand command, rawCallback callback);
  /** Start active phase. This function tells the game that we have
      finished setting up commands and channels, and are ready to start
      sending and receiving game data.
  */
  void activate();

  /* Active phase */
  /** Send a command to the game. See the protocol documentation for a
      full list of possible commands.
      @param command The command constant to send.
  */
  void sendCommand(OISCommand command);
  /** Send a log message to the game. The game will treat this as a debug
      log, and add it to the game log file.
      @param message A string to be logged.
  */
  void sendDebug(char* message);
  /** Helper function to convert data received from the game to
      a floating point number.
  */
  float toFloat(int data);
  /** The main ObjectsInSpace loop. Call this regularly in your loop()
      function.
  */
  void update();
  #ifdef OIS_DEBUG
  /** An optional callback to receive error messages. To use this, add
      `#define OIS_DEBUG` to the top of your sketch, and create an
      OISErrorCallback function that will receive and process errors.
  */
  void errorHandler(OISErrorCallback callback);
  #endif
 private:
  // Stores the maximum number of request channels
  int _numRequests;
  // Stores the maximum number of managed input pins
  int _numCommandPins;

  // Our serial device
  Stream *_serial;

  OISState _streamState;
  // This will hold an array of callback functions
  AISRequest** _commandCallbacks;
  // This will hold an array of managed input commands
  AISManagedInputPin** _managedInputPins;

  // Tracks _commandCallbacks index while filling the array in Syncing phase.
  int _callbackidx;
  // Tracks _managedInputPins index while filling the array in Syncing phase.
  int _inputidx;

  // Serial buffers
  char _sendBuffer[sendBufSize];
  char _recvBuffer[recvBufSize];
  // Tracks current position reading in to _recvBuffer
  int _recvIdx;

  bool _validateManagedInputs(OISCommand firstCommand,
			      OISCommand secondCommand);
  bool _validateCommand(OISCommand command);
  void _sendRequest(OISRequest request, OISCommand command);

  void _parseSerial();


  #ifdef OIS_DEBUG
  OISErrorCallback _errorCallback;
  #endif
};

#endif
